﻿//+--------------------------------------------------------------------------+
//|                                                                          |
//|                                 glChart                                  |
//|                              Version 1.00                                |
//|                                                                          |
//|      OpenGL project using the GDImage flat API to create on the fly      |
//|     textured 3D tubular chart with smooth shading and ambient light.     |
//|                                                                          |
//|          The GDImage.dll unregistered version has no limitation,         |
//|             and all features of the API are fully functional.            |
//|                                                                          |
//|       Buying a commercial license removes the copyright nag screen.      |
//|                                                                          |
//+--------------------------------------------------------------------------+
//|                                                                          |
//|                         Author Patrice TERRIER                           |
//|                           copyright (c) 2006                             |
//|                                                                          |
//|                        pterrier@zapsolution.com                          |
//|                                                                          |
//|                          www.zapsolution.com                             |
//|                                                                          |
//+--------------------------------------------------------------------------+
//|                  Project started on : 11-15-2006 (MM-DD-YYYY)            |
//|                        Last revised : 11-17-2006 (MM-DD-YYYY)            |
//+--------------------------------------------------------------------------+

using System;
//using System.Collections.Generic;
//using System.ComponentModel;
//using System.Data;
using System.Drawing;
//using System.Text;
using System.Windows.Forms;

using System.Runtime.InteropServices;
//using System.Runtime;
//using System.IO;
//using Microsoft.Win32;

using Win32;
using GDImage;
using OpenGL;
using SkinEngine;

namespace glChart
{
    //___________________________________________________________
    public delegate int CallBack(IntPtr hWnd, int wMsg, uint wParam, uint lParam);
    //___________________________________________________________

    public partial class MAIN_Form : Form
    {
        private struct THISCHART
        {
            public int chartnumber;
            public int chartseries;
            public float radius;
            //public float barsize;
            public float xoff;
            public float yoff;
            public float zoff;
            public float xrot;
            public float yrot;
            public float xcpy;
            public float ycpy;
            //public int move;
            public int xdown;
            public int ydown;
            //public int mouseDown;
            public int listIndex;
        }

        //_______________________________________________
        const int T1      = 0;                         // Texture 1
        const int T2      = 1;                         // Texture 2
        const int T3      = 2;                         // Texture 3
        const int T4      = 3;                         // Texture 4
        const int BACK    = 4;                         // Texture 5
        const int T6      = 5;                         // BACK texture
        const int ID_CTRL = 100;                       // GDImage Zoom control
        const int ID_T1   = 101;
        const int ID_T2   = 102;
        const int ID_T3   = 103;
        const int ID_T4   = 104;
        const int ID_T5   = 105;
        const int ID_BACK = 106;
        const int TIMER_MAX = 2;                       // To perform smooth animation

        const int chartseries = 7;
        const int chartnumber = 4;

        static THISCHART chart;
        static int quadObj;
        static int[] thisChartItem = new int[(chartnumber * chartseries) + 1];

        static GI.ZGLTEXTURE[] mt = new GI.ZGLTEXTURE[6];

        static public IntPtr hFORM_Main = IntPtr.Zero; // Default to Zero
        static public IntPtr hGLcontrol = IntPtr.Zero; // Default to Zero

        static GI.ZGLFONT UseFont = new GI.ZGLFONT();

        static int WasZoom = 0;
        static bool Zooming = false;

        //____________________________________________________________
        [DllImport("USER32", EntryPoint = "SetWindowLongA")]
        public static extern int SetCallBack(IntPtr hWnd, int nIndex, CallBack lNewLong);
        static CallBack UseCallBack = null;
        static int ghGLproc;                             // OpenGL subclassing

        static int wasLX, wasLY, wasRX, wasRY;
        static bool LeftMousing, RightMousing;
        //____________________________________________________________

        public MAIN_Form()
        {
            InitializeComponent();

            // Create the OpenGL control
            hGLcontrol = GI.CreateGLcontrol(GL_Template, ID_CTRL);

            hFORM_Main = this.Handle;
            //StartAnimation();
        }

        private void FORM_Load(object sender, EventArgs e)
        {
            // Set default image folder
            GI.LoadSavePath(Api.ExePath(), true);

            // We need this to retrieve bitmap from resource
            SK.UseNameSpace = this.GetType().Namespace;

            if (hGLcontrol != IntPtr.Zero)
            {
                //______________________________________________________________
                UseCallBack = new CallBack(glWndProc);
                ghGLproc = SetCallBack(hGLcontrol, Api.GWL_WNDPROC, UseCallBack);
                //______________________________________________________________

                OnOff.Checked = true;
                Texture.Checked = true;

                // Do not fit the texture into a square shape
                GI.DoNotSquareTexture();

                mt[0].FullName = Api.ExePath() + "Plasma.jpg";   //mt[T1].ID = ID_T1;
                mt[1].FullName = Api.ExePath() + "Oblique.jpg";  //mt[T2].ID = ID_T2;
                mt[2].FullName = Api.ExePath() + "Nebulous.jpg"; //mt[T3].ID = ID_T3;
                mt[3].FullName = Api.ExePath() + "Water.jpg";    //mt[T4].ID = ID_T4;
                mt[BACK].FullName = Api.ExePath() + "Back.jpg";  //mt[BACK].ID = ID_BACK;
                mt[5].FullName = Api.ExePath() + "Cover.png";    //mt[T5].ID = ID_T5;

                // Assign textures to mt[]
                SetGenTextures(mt.Length);

                // I am using a temporary ZGLTEXTURE structure 
                // to protect mt[] from being erased during 
                // call to SetMutipleGLTextureFrom32Bit.
                GI.ZGLTEXTURE[] temp = new GI.ZGLTEXTURE[mt.Length];
                for (int k = 0; k < mt.Length; k++) { temp[k] = mt[k]; }
                if (GI.CreateMutipleGLtextures(ref temp, mt.Length) == 0)
                {
                    InitGraph();
                    GL.glEnable(GL.GL_TEXTURE_2D);

                    // Setup drawing mode
                       //GL.glClearColor(0.9f, 0.9f, 1f, 1f);
                       //GL.glClearDepth(1.0f);
                    GL.glDepthFunc(GL.GL_LESS);
                    GL.glEnable(GL.GL_DEPTH_TEST);
                    GL.glShadeModel(GL.GL_SMOOTH);
                    GL.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);

                    // Create quadric object
                    quadObj = GL.gluNewQuadric();
                    GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
                    GL.glEnable(GL.GL_BLEND);
                    GL.glEnable(GL.GL_LINE_SMOOTH);
                    GL.gluQuadricNormals(quadObj, GL.GLU_SMOOTH);

                    // Setup ambient lighting
                    float[] LightDiffuse = { 1.0f, 1.0f, 1.0f, 1.0f };
                    float[] LightAmbient = { 0.05f, 0.05f, 0.05f, 1.0f };
                    float[] LightPosition = { -50.0f, 50.0f, 100.0f, 1.0f };
                    GL.glLightfv(GL.GL_LIGHT1, GL.GL_DIFFUSE, LightDiffuse);     //' Setup The Diffuse Light
                    GL.glLightfv(GL.GL_LIGHT1, GL.GL_AMBIENT, LightAmbient);     //' Setup The Ambient Light
                    GL.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, LightPosition);   //' Position The Light
                    GL.glEnable(GL.GL_LIGHT1);                                   //' Enable Light ONE
                    GL.glEnable(GL.GL_LIGHTING);                                 //' Enable Lighting
                    GL.glEnable(GL.GL_COLOR_MATERIAL);                           //' Enable Coloring Of Material
                    GL.glColorMaterial(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE);

                    // Init list index
                    chart.listIndex = 1;
                    GL.glGenLists(chart.listIndex);
                    // Create x,y,z axis
                    float yRange = 9.0f; // including the bootom line
                    DrawAxis(15.0f, yRange, 11.0f);
                    
                    float TheValue; Random rnd = new Random(); // To compute random value     
                    int k = 0;
                    for (int i = 1; i < chart.chartnumber + 1; i++)
                    {
                        for (int j = 1; j < chart.chartseries + 1; j++)
                        {
                            TheValue = rnd.Next(1, (int)yRange - 1); // compute fictive height value
                            k++; thisChartItem[k] = chart.listIndex + k;
                            GL.glNewList(thisChartItem[k], GL.GL_COMPILE);
                               GL.glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);

                               GL.glBindTexture(GL.GL_TEXTURE_2D, i);
                               GL.gluQuadricNormals(quadObj, GL.GLU_SMOOTH); // Create Smooth Normals
                               GL.gluQuadricTexture(quadObj, 1);             // Create Texture Coords
                               GL.gluCylinder(quadObj, chart.radius, chart.radius, TheValue, 32, 1);

                               GL.glBindTexture(GL.GL_TEXTURE_2D, 6);
                               GL.gluDisk(quadObj, 0.0f, chart.radius, 32, 32);
                               GL.glTranslatef(0.0f, 0.0f, TheValue);
                               GL.gluDisk(quadObj, 0.0f, chart.radius, 32, 32);
                               GL.glTranslatef(0.0f, 0.0f, -TheValue);
                               GL.glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
                            GL.glEndList();
                        }
                    }

                    // Create GDImage GL font
                    UseFont.fontName = "Arial";       // A True Type font
                    UseFont.fontHeight = 10;          // Size of the font
                    UseFont.fontWeight = Api.FW_BOLD; // Font style
                    GI.BuildGLfont(GI.GetGLDC(hGLcontrol), ref UseFont);

                    // When using the GDImage file selector
                    // start looking for pictures into the EXE folder.
                    GI.LoadSavePath(Api.ExePath(), true);
                }
            }
        }

        private void SetGenTextures(int n)
        {
            int[] textures = new int[n];
            GL.glGenTextures(n, textures);
            for (int k = 0; k < mt.Length; k++) { mt[k].Texture = textures[k]; }
        }

        private void DrawAxis(float x, float y, float z)
        {
            float[] COLOR1 = { 0.6f, 0.6f, 0.6f, 0.3f };
            float[] COLOR2 = { 1.0f, 1.0f, 1.0f, 1.0f };
            float[] COLOR3 = { 0.6f, 0.0f, 0.0f, 1.0f };

            GL.glNewList(chart.listIndex, GL.GL_COMPILE);
         
                GL.glBegin(GL.GL_QUADS);
                    GL.glColor4fv(COLOR1);
                    GL.glVertex3f(0f, y, z);
                    GL.glVertex3f(0f, -1f, z);
                    GL.glVertex3f(0f, -1f, -1f);
                    GL.glVertex3f(0f, y, -1f);

                    GL.glVertex3f(-1f, y, 0f);
                    GL.glVertex3f(-1f, -1f, 0f);
                    GL.glVertex3f(x, -1f, 0f);
                    GL.glVertex3f(x, y, 0f);
                GL.glEnd();

                GL.glColor4fv(COLOR2);
                for (int a = 1; a < (int)y; a++)
                {
                    GL.glBegin(GL.GL_LINE_STRIP);
                       GL.glVertex3f(0.1f, (float)a, z);
                       GL.glVertex3f(0.1f, (float)a, 0.1f);
                       GL.glVertex3f(x, (float)a, 0.1f);
                    GL.glEnd();
                }

                GL.glColor4fv(COLOR3);
                GL.glBegin(GL.GL_LINE_STRIP);
                   GL.glVertex3f(0.1f, 0f, z);
                   GL.glVertex3f(0.1f, 0f, 0.1f);
                   GL.glVertex3f(x, 0f, 0.1f);
                GL.glEnd();
                GL.glBegin(GL.GL_LINES);
                   GL.glVertex3f(0.1f, -1f, 0.1f);
                   GL.glVertex3f(0.1f, y, 0.1f);
                GL.glEnd();


            GL.glEndList();
        }

        //private void StartAnimation()
        //{
        //    // Use API WM_TIMER to avoid aggressive garbage collection 
        //    // to occur before Main ends.
        //    for (int T = 0; T < TIMER_MAX; T++)
        //    {
        //        Api.SetTimer(hFORM_Main, T + 1, 0, null);
        //    }
        //}

        private void SetZoom(int zoom)
        {
            GI.SetGLzoom(hGLcontrol, zoom);
            GI.ResizeGLWindow(hGLcontrol);
        }

        private void UpdateZoom(int z)
        {
            SetZoom(z); RenderOpenGL(hGLcontrol);
        }

        protected override void WndProc(ref Message m)
        {
            switch (m.Msg)
            {
                //case Api.WM_TIMER:
                //    RenderOpenGL(hGLcontrol);
                //    m.Result = (IntPtr)0;
                //    return;

                case Api.WM_COMMAND:
                    if (m.LParam == hGLcontrol)
                    {
                        RenderOpenGL(hGLcontrol);
                        break;
                    }
                    
                    //  Zooming part
                    if (m.LParam == ZOOM_Track.Handle)
                    {// So much easier than trying to raise an event
                        // Warning, we are using polar coordinates there
                        // thus convert it to Degree
                        if (Zooming == false)
                        {
                            Zooming = true;
                            double range = ZOOM_Track.Maximum - ZOOM_Track.Minimum;
                            double increment = range / 90.0f;
                            int ZoomIs = (int)Math.Max(range - (ZOOM_Track.Value * increment), 1);

                            if (WasZoom < ZoomIs)
                            { for (int z = WasZoom; z < ZoomIs + 1; z++) UpdateZoom(z); }
                            else if (WasZoom > ZoomIs)
                            { for (int z = WasZoom; z > ZoomIs - 1; z--) UpdateZoom(z); }

                            WasZoom = ZoomIs;
                            Zooming = false;
                        }
                    }
                    break;

                case Api.WM_DESTROY:
                    DisposeGL(); // Free up OpenGL resources
                    break;
            }

            base.WndProc(ref m);

            if (m.Msg == Api.WM_COMMAND) Api.SetFocus(m.HWnd);            
        }

        private void Form_PreviewKeydown(object sender, PreviewKeyDownEventArgs e)
        {
            int ZoomIs;
            switch ((int)e.KeyCode)
            {
                case Api.VK_LEFT: chart.xoff -= 0.05f; RenderOpenGL(hGLcontrol); break;
                case Api.VK_RIGHT: chart.xoff += 0.05f; RenderOpenGL(hGLcontrol); break;
                case Api.VK_UP: chart.yoff -= 0.05f; RenderOpenGL(hGLcontrol); break;
                case Api.VK_DOWN: chart.yoff += 0.05f; RenderOpenGL(hGLcontrol); break;
                case Api.VK_PGUP:
                    ZoomIs = GI.GetGLzoom(hGLcontrol);
                    if (ZoomIs > 1)
                    {
                        GI.SetGLzoom(hGLcontrol, ZoomIs - 1);
                        GI.ResizeGLWindow(hGLcontrol);
                        RenderOpenGL(hGLcontrol);
                    }
                    break;
                case Api.VK_PGDN:
                    ZoomIs = GI.GetGLzoom(hGLcontrol);
                    if (ZoomIs < 100)
                    {
                        GI.SetGLzoom(hGLcontrol, ZoomIs + 1);
                        GI.ResizeGLWindow(hGLcontrol);
                        RenderOpenGL(hGLcontrol);
                    }
                    break;
            }
        }

        // Free up resources
        private void DisposeGL()
        {
            //for (int T = 0; T < TIMER_MAX; T++)
            //{
            //    Api.KillTimer(hFORM_Main, T + 1);
            //}

            // Delete OpenGL font.
            GI.DeleteGLFont(ref UseFont);
            // Delete quadObject
            if (quadObj != 0) GL.gluDeleteQuadric(quadObj); quadObj = 0;
        }

        private void RenderOpenGL(IntPtr hContainer)
        {
            // Save the active planet zoom, 
            // then switch to a fixed zoom for the background
            int ZoomIs = GI.GetGLzoom(hGLcontrol);
            SetZoom(64);

            // Switch back to default color
            GL.glColor3f(1f, 1f, 1f); 

            GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
            GL.glLoadIdentity();

            // ***************
            // Draw background
            // ***************
            GL.glBindTexture(GL.GL_TEXTURE_2D, mt[BACK].Texture);
            // Disable Depth Test
            GL.glDisable(GL.GL_DEPTH_TEST);
            GL.glBegin(GL.GL_QUADS);
               GL.glTexCoord2f(0.0f, 0.0f); GL.glVertex3f(-60.0f, -50.0f, -80.0f);// Bottom left of texture and quad
               GL.glTexCoord2f(1.0f, 0.0f); GL.glVertex3f(60.0f, -50.0f, -80.0f); // Bottom right
               GL.glTexCoord2f(1.0f, 1.0f); GL.glVertex3f(60.0f, 50.0f, -80.0f);  // Top right
               GL.glTexCoord2f(0.0f, 1.0f); GL.glVertex3f(-60.0f, 50.0f, -80.0f); // Top left
            GL.glEnd();
            // ***********************

            // Restore the previously saved Zoom to display the chart
            SetZoom(ZoomIs);

            // ***************
            // Draw chart
            // ***************
            // Enable Depth Test
            GL.glEnable(GL.GL_DEPTH_TEST);

            GL.glTranslatef(chart.xoff, chart.yoff, chart.zoff);

            GL.glRotatef(chart.xrot, 1.0f, 0.0f, 0.0f);
            GL.glRotatef(chart.yrot, 0.0f, 1.0f, 0.0f);

            // Draw the XYZ grid
            GL.glLineWidth(1.0f);
            GL.glCallList(chart.listIndex);

            // Shall we use the texture mode
            if (!Texture.Checked) GL.glDisable(GL.GL_TEXTURE_2D);
            int k = 0;
            GL.glTranslatef(chart.radius + 0.5f, 0f, chart.radius + 0.5f);
            for (int i = 1; i < chart.chartnumber + 1; i++)
            {// Set color for the chart serie
                switch (i)
                {
                    case 1: GI.glColor4f(Api.ARGB(255, 255, 255, 255)); break;
                    case 2: GI.glColor4f(Api.ARGB(220, 0, 255, 0)); break;
                    case 3: GI.glColor4f(Api.ARGB(200, 0, 0, 255)); break;
                    case 4: GI.glColor4f(Api.ARGB(192, 255, 0, 0)); break;
                }
                for (int j = 1; j < chart.chartseries + 1; j++)
                {// Draw the matching chart item
                    k++; GL.glCallList(thisChartItem[k]);
                    GL.glTranslatef(2.0f * chart.radius, 0f, 0f);
                }
                GL.glTranslatef(-2.0f * chart.radius * chart.chartseries, 0.0f, 2.0f * chart.radius + 0.5f);
            }
            // Make sure to restore the texture mode
            if (!Texture.Checked) GL.glEnable(GL.GL_TEXTURE_2D);

            // ***********************
            // Display coordinates
            // ***********************
            if (OnOff.Checked)
            {
                int UseColor = Api.ARGB(255, 247, 225, 125);
                string MyStr = "Horizontal " + (chart.xoff + 3.5).ToString("0.##");
                GI.DrawGLText(hContainer, ref UseFont, 9, 10, MyStr, UseColor);
                MyStr = "Vertical " + (chart.yoff - 3.25).ToString("0.##");
                GI.DrawGLText(hContainer, ref UseFont, 25, 10 + UseFont.fontHeight, MyStr, UseColor);
                MyStr = "Zoom " + ((int)ZOOM_Track.Value).ToString();
                GI.DrawGLText(hContainer, ref UseFont, 39, 10 + UseFont.fontHeight * 2, MyStr, UseColor);
            }
            // ***********************

            // ***********************
            // Draw the scene
            // ***********************
            GI.UpdateGLWindow(hContainer);
            // ***********************
        }

        private void InitGraph()
        {
            // Graph Offsets
            chart.xoff = -3.5f;
            chart.yoff = 3.25f;
            chart.zoff = -36;

            // Graph Rotation
            chart.xrot = 45.0f;
            chart.yrot = -30.0f;

            chart.radius = 1.0f;

            chart.chartseries = chartseries;
            chart.chartnumber = chartnumber;

            ZOOM_Track.Value = WasZoom = 50;
        }

        // ****************************************
        private void FORM_Close(object sender, EventArgs e)
        {
        }

        //***********************************************************
        private void FORM_Resize(object sender, EventArgs e)
        {// Use the template to resize the GDImage OpenGL control
            Api.MoveWindow(hGLcontrol, GL_Template.Location.X, GL_Template.Location.Y, GL_Template.Width, GL_Template.Height, true);
            GI.ResizeGLWindow(hGLcontrol);
            RenderOpenGL(hGLcontrol);
        }

        private void BTN_Select_Click(object sender, EventArgs e)
        {
            string sFileName = GI.LoadDialog(hFORM_Main);
            if (sFileName != null)
            {
                //GI.DoNotSquareTexture();
                int NamedTexture = mt[BACK].Texture;
                GI.UpdateNamedGLTextureFromFile(sFileName, NamedTexture);
            }
        }

        private void BTN_Reset_Click(object sender, EventArgs e)
        {
            InitGraph();
            GI.SetGLzoom(hGLcontrol, 0); GI.ResizeGLWindow(hGLcontrol);
            RenderOpenGL(hGLcontrol);
        }

        private void OnOff_CheckedChanged(object sender, EventArgs e)
        {
            RenderOpenGL(hGLcontrol);
        }

        private void Texture_CheckedChanged(object sender, EventArgs e)
        {
            RenderOpenGL(hGLcontrol);
        }

        static void RedrawGLcontrol()
        {// I am doing this to bypass the use of a timer
            Api.SendMessage(hFORM_Main, Api.WM_COMMAND, (uint)hGLcontrol, (int)hGLcontrol);
        }

        //_________________________________________________________________________
        public static int glWndProc(IntPtr hWnd, int wMsg, uint wParam, uint lParam)
        {
            int x, y;
            switch (wMsg)
            {
                case Api.WM_PAINT:
                    RedrawGLcontrol();
                    break;
                case Api.WM_LBUTTONDOWN:
                    if (LeftMousing == false)
                    {
                        x = Api.LoWrd(lParam); y = Api.HiWrd(lParam);
                        chart.xcpy = chart.xrot;
                        chart.ycpy = chart.yrot;
                        //CALL SetClassLong(glWnd, %GCL_HCURSOR, LoadCursor(0, BYVAL %IDC_SIZEALL))
                        chart.xdown = x;
                        chart.ydown = y;
                        LeftMousing = true;
                    }
                    if (Api.GetFocus() != hWnd) Api.SetFocus(hWnd);
                    break;
                case Api.WM_LBUTTONUP:
                    LeftMousing = false;
                    //CALL SetClassLong(glWnd, %GCL_HCURSOR, LoadCursor(0, BYVAL %IDC_ARROW))
                    break;
                case Api.WM_RBUTTONDOWN:
                    if (RightMousing == false)
                    {
                        x = Api.LoWrd(lParam); y = Api.HiWrd(lParam);
                        chart.xcpy = chart.xoff;
                        chart.ycpy = chart.yoff;
                        //CALL SetClassLong(glWnd, %GCL_HCURSOR, LoadCursor(0,  BYVAL %IDC_HAND))
                        chart.xdown = x;
                        chart.ydown = y;
                        RightMousing = true;
                    }
                    if (Api.GetFocus() != hWnd) Api.SetFocus(hWnd);
                    break;
                case Api.WM_RBUTTONUP:
                    RightMousing = false;
                    //CALL SetClassLong(glWnd, %GCL_HCURSOR, LoadCursor(0, BYVAL %IDC_ARROW))
                    break;
                case Api.WM_MOUSEMOVE:
                    if (!GI.IsLButtonDown())
                    {
                        LeftMousing = false;
                        if (RightMousing == false)
                        {
                            //CALL SetClassLong(glWnd, %GCL_HCURSOR, LoadCursor(0, BYVAL %IDC_ARROW))
                        }
                    }
                    else
                    {
                        x = Api.LoWrd(lParam); y = Api.HiWrd(lParam);
                        if ((wasLX != x) || (wasLY != y))
                        {
                            if (LeftMousing)
                            {
                               int dx = x - chart.xdown; chart.yrot = chart.ycpy + dx / 2.0f;
                               int dy = y - chart.ydown; chart.xrot = chart.xcpy + dy / 2.0f;
                            }
                        }
                        wasLX = x; wasLY = y;
                        RedrawGLcontrol(); ;
                        break;
                    }

                    if (!GI.IsRButtonDown())
                    {
                        RightMousing = false;
                        //CALL SetClassLong(glWnd, %GCL_HCURSOR, LoadCursor(0, BYVAL %IDC_ARROW))
                    }
                    else
                    {
                        x = Api.LoWrd(lParam); y = Api.HiWrd(lParam);
                        if ((wasLX != x) || (wasLY != y))
                        {
                            if (RightMousing)
                            {
                                Api.RECT rc = new Api.RECT();
                                Api.GetClientRect(hGLcontrol, ref rc);
                                int dx = x - chart.xdown; chart.xoff = chart.xcpy - (-45 * dx) / (2.0f * rc.right);
                                int dy = y - chart.ydown; chart.yoff = chart.ycpy + (-45 * dy) / (2.0f * rc.bottom);
                            }
                            RedrawGLcontrol(); ;
                            break;
                        }
                        wasRX = x; wasRY = y;
                    }
                    RedrawGLcontrol(); ;
                    break;
            }
            return Api.CallWindowProc(ghGLproc, hWnd, wMsg, wParam, lParam);
        }

        //_________________________________________________________________________

    }
}